(*
Author: Robert Sison
*)

theory RegsSimple
imports "HOL-Word.Word"
begin

(* Naive for initial instantiation: 32 GPRs *)
datatype reg =
  r0 |
  r1 |
  r2 |
  r3 |
  r4 |
  r5 |
  r6 |
  r7 |
  r8 |
  r9 |
  r10 |
  r11 |
  r12 |
  r13 |
  r14 |
  r15 |
  r16 |
  r17 |
  r18 |
  r19 |
  r20 |
  r21 |
  r22 |
  r23 |
  r24 |
  r25 |
  r26 |
  r27 |
  r28 |
  r29 |
  r30 |
  r31

instantiation reg :: "{enum, linorder}"
begin

abbreviation
  "reg_enum \<equiv> [
    r0,
    r1,
    r2,
    r3,
    r4,
    r5,
    r6,
    r7,
    r8,
    r9,
    r10,
    r11,
    r12,
    r13,
    r14,
    r15,
    r16,
    r17,
    r18,
    r19,
    r20,
    r21,
    r22,
    r23,
    r24,
    r25,
    r26,
    r27,
    r28,
    r29,
    r30,
    r31
    ]"

definition enum_reg_def:
  "enum_class.enum \<equiv> reg_enum"

primrec index_of_reg :: "reg \<Rightarrow> nat"
where
  "index_of_reg r0 = 0" |
  "index_of_reg r1 = 1" |
  "index_of_reg r2 = 2" |
  "index_of_reg r3 = 3" |
  "index_of_reg r4 = 4" |
  "index_of_reg r5 = 5" |
  "index_of_reg r6 = 6" |
  "index_of_reg r7 = 7" |
  "index_of_reg r8 = 8" |
  "index_of_reg r9 = 9" |
  "index_of_reg r10 = 10" |
  "index_of_reg r11 = 11" |
  "index_of_reg r12 = 12" |
  "index_of_reg r13 = 13" |
  "index_of_reg r14 = 14" |
  "index_of_reg r15 = 15" |
  "index_of_reg r16 = 16" |
  "index_of_reg r17 = 17" |
  "index_of_reg r18 = 18" |
  "index_of_reg r19 = 19" |
  "index_of_reg r20 = 20" |
  "index_of_reg r21 = 21" |
  "index_of_reg r22 = 22" |
  "index_of_reg r23 = 23" |
  "index_of_reg r24 = 24" |
  "index_of_reg r25 = 25" |
  "index_of_reg r26 = 26" |
  "index_of_reg r27 = 27" |
  "index_of_reg r28 = 28" |
  "index_of_reg r29 = 29" |
  "index_of_reg r30 = 30" |
  "index_of_reg r31 = 31"

definition enum_reg_all_def: "enum_class.enum_all P \<equiv> \<forall>r \<in> set reg_enum. P r"
definition enum_reg_ex_def: "enum_class.enum_ex P \<equiv> \<exists>r \<in> set reg_enum. P r"

definition linorder_reg_less_eq_def: "ord_class.less_eq a b \<equiv> index_of_reg a \<le> index_of_reg b"
definition linorder_reg_less_def: "ord_class.less a b \<equiv> index_of_reg a < index_of_reg b"

instance proof
  show "(UNIV::reg set) = set enum_class.enum"
  unfolding enum_reg_def
  apply(clarsimp | safe)+
  by (rename_tac x, case_tac x, auto)
next
  show "distinct (enum_class.enum::reg list)"
  unfolding enum_reg_def by simp
next
  show "\<And>P::reg \<Rightarrow> bool. enum_class.enum_all P = Ball UNIV P"
  unfolding enum_reg_all_def
  apply(clarsimp | safe)+
  by (rename_tac x, case_tac x, auto)
next
  show "\<And>P::reg \<Rightarrow> bool. enum_class.enum_ex P = Bex UNIV P"
  unfolding enum_reg_ex_def
  by (auto, rename_tac x, case_tac x, auto)
next
  show "\<And>(x::reg) y::reg. (x < y) = (x \<le> y \<and> \<not> y \<le> x)"
  unfolding linorder_reg_less_eq_def linorder_reg_less_def by linarith
next
  show "\<And>x::reg. x \<le> x"
  unfolding linorder_reg_less_eq_def by linarith
next
  show "\<And>(x::reg) (y::reg) z::reg. x \<le> y \<Longrightarrow> y \<le> z \<Longrightarrow> x \<le> z"
  unfolding linorder_reg_less_eq_def by linarith
next
  show "\<And>(x::reg) y::reg. x \<le> y \<Longrightarrow> y \<le> x \<Longrightarrow> x = y"
  unfolding linorder_reg_less_eq_def
  by (case_tac x, (case_tac y, auto)+)
next
  show "\<And>(x::reg) y::reg. x \<le> y \<or> y \<le> x"
  unfolding linorder_reg_less_eq_def by linarith
qed

end

end
